@@ -2477,8 +2477,13 @@ pcre_compile_nuls (pattern, patlen, options,
 	    length++;
 	  continue;
 
-	case '^':
 	case '.':
+	  if (ptr != compile_block.end && *(ptr + 1) == '{')
+	    length++;
+
+	  /* fall through */
+
+	case '^':
 	case '$':
 	case '*':		/* These repeats won't be after brackets; */
 	case '+':		/* those are handled separately */
@@ -3008,8 +3013,7 @@ pcre_compile_nuls (pattern, patlen, options,
 
 #ifndef DEBUG
   if (code - re->code > length)
-    printf ("%d %d\n", code - re->code, length);
-  //    *errorptr = pcre_estrings[23];
+      *errorptr = pcre_estrings[23];
 #endif
 
 /* Give an error if there's back reference to a non-existent capturing
