@@ -2416,7 +2416,7 @@ pcre_posix_compile_nuls (pattern, patlen, options,
 	    goto NORMAL_CHAR;
 
 	case '.':
-	  length++;
+          length += (ptr != compile_block.end && *(ptr + 1) == '{') ? 2 : 1;
 	  continue;
 
 	  /* This covers the cases of repeats after a single char, metachar, class,
