@@ -54,6 +54,14 @@ extern int errno;
 # include <memory.h>
 #endif
 
+
+#if defined HAVE_WCTYPE_H && defined HAVE_WCHAR_H && defined HAVE_MBRTOWC
+/* We can handle multibyte string.  */
+# include <wchar.h>
+# include <wctype.h>
+# define MBS_SUPPORT
+#endif
+
 #ifndef HAVE_STRCHR
 # define strchr index
 # define strrchr rindex
@@ -76,18 +84,6 @@ extern int errno;
 #include "utils.h"
 #include "sed.h"
 
-/* If set, fflush(stdout) on every line output. */
-extern flagT unbuffered_output;
-
-/* If set, don't write out the line unless explicitly told to. */
-extern flagT no_default_output;
-
-/* Do we need to be pedantically POSIX compliant? */
-extern flagT POSIXLY_CORRECT;
-
-/* How long should the `l' command's output line be? */
-extern countT lcmd_out_line_len;
-
 
 /* Sed operates a line at a time. */
 struct line {
@@ -99,12 +95,13 @@ struct line {
 };
 
 /* A queue of text to write out at the end of a cycle
-   (filled by the "a" and "r" commands.) */
+   (filled by the "a", "r" and "R" commands.) */
 struct append_queue {
-  const char *rfile;
-  const char *text;
+  const char *fname;
+  char *text;
   size_t textlen;
   struct append_queue *next;
+  flagT free;
 };
 
 /* State information for the input stream. */
@@ -404,10 +401,11 @@ next_append_slot()
 {
   struct append_queue *n = MALLOC(1, struct append_queue);
 
-  n->rfile = NULL;
+  n->fname = NULL;
   n->text = NULL;
   n->textlen = 0;
   n->next = NULL;
+  n->free = 0;
 
   if (append_tail)
       append_tail->next = n;
@@ -424,6 +422,9 @@ release_append_queue()
 
   for (p=append_head; p; p=q)
     {
+      if (p->free)
+        FREE(p->text);
+
       q = p->next;
       FREE(p);
     }
@@ -440,14 +441,14 @@ dump_append_queue()
     {
       if (p->text)
 	  output_line(p->text, p->textlen, 0, output_file);
-      if (p->rfile)
+      if (p->fname)
 	{
 	  char buf[FREAD_BUFFER_SIZE];
 	  size_t cnt;
 	  FILE *fp;
 
-	  fp = fopen(p->rfile, "r");
-	  /* Not ck_fopen() because: "If _rfile_ does not exist or cannot be
+	  fp = fopen(p->fname, "r");
+	  /* Not ck_fopen() because: "If _fname_ does not exist or cannot be
 	     read, it shall be treated as if it were an empty file, causing
 	     no error condition."  IEEE Std 1003.2-1992 */
 	  if (fp)
@@ -528,13 +529,9 @@ open_next_file(name, input)
 
       /* get the base name */
       if (p = strrchr(tmpdir, '/'))
-        {
-	    *(p + 1) = 0;
-	}
+	*(p + 1) = 0;
       else
-        {
-	    strcpy(tmpdir, ".");
-	}
+	strcpy(tmpdir, ".");
       
       input->in_file_name = name;
       input->out_file_name = temp_file_template (tmpdir, "sed");
@@ -564,29 +561,46 @@ closedown(input)
   if (input->fp != stdin) /* stdin can be reused on tty and tape devices */
     ck_fclose(input->fp);
 
-  if (output_file != stdout)
-    {
-      if (strcmp(in_place_extension, "*") != 0)
+      if (in_place_extension && output_file != NULL)
 	{
-	  char *backup_file_name = get_backup_file_name(input->in_file_name);
-	  rename (input->in_file_name, backup_file_name);
-	  free (backup_file_name);
+	  if (strcmp(in_place_extension, "*") != 0)
+	    {
+	      char *backup_file_name = get_backup_file_name(input->in_file_name);
+	      rename (input->in_file_name, backup_file_name);
+	      free (backup_file_name);
+	    }
+
+	  ck_fclose (output_file);
+	  rename (input->out_file_name, input->in_file_name);
+	  free (input->out_file_name);
 	}
 
-      fclose (output_file);
-      rename (input->out_file_name, input->in_file_name);
-      free (input->out_file_name);
-    }
-
   input->fp = NULL;
+
+  if (separate_files)
+    rewind_read_files ();
+}
+
+/* Reset range commands so that they are marked as non-matching */
+static void reset_addresses P_((struct vector *));
+static void
+reset_addresses(vec)
+     struct vector *vec;
+{
+  struct sed_cmd *cur_cmd;
+  int n;
+
+  for (cur_cmd = vec->v, n = vec->v_length; n--; cur_cmd++)
+    cur_cmd->a1_matched = 0;
 }
 
 /* Read in the next line of input, and store it in the pattern space.
    Return zero if there is nothing left to input. */
-static flagT read_pattern_space P_((struct input *, flagT append));
+static flagT read_pattern_space P_((struct input *, struct vector *, flagT));
 static flagT
-read_pattern_space(input, append)
+read_pattern_space(input, the_program, append)
   struct input *input;
+  struct vector *the_program;
   flagT append;
 {
   if (append_head) /* redundant test to optimize for common case */
@@ -599,12 +613,20 @@ read_pattern_space(input, append)
   while ( ! (*input->read_fn)(input) )
     {
       closedown(input);
+
       if (!*input->file_list)
 	{
 	  line.chomped = 0;
 	  return 0;
 	}
-      open_next_file(*input->file_list++, input);
+
+      if (separate_files)
+	{
+	  input->line_number = 0;
+	  reset_addresses (the_program);
+	}
+
+      open_next_file (*input->file_list++, input);
     }
 
   ++input->line_number;
@@ -637,9 +659,9 @@ last_file_with_data_p(input)
 }
 
 /* Determine if we match the `$' address. */
-static flagT test_dollar_EOF P_((struct input *));
+static flagT test_eof P_((struct input *));
 static flagT
-test_dollar_EOF(input)
+test_eof(input)
   struct input *input;
 {
   int ch;
@@ -647,11 +669,11 @@ test_dollar_EOF(input)
   if (buffer.length)
     return 0;
   if (!input->fp)
-    return last_file_with_data_p(input);
+    return separate_files || last_file_with_data_p(input);
   if (feof(input->fp))
-    return last_file_with_data_p(input);
+    return separate_files || last_file_with_data_p(input);
   if ((ch = getc(input->fp)) == EOF)
-    return last_file_with_data_p(input);
+    return separate_files || last_file_with_data_p(input);
   ungetc(ch, input->fp);
   return 0;
 }
@@ -694,7 +716,7 @@ match_an_address_p(addr, input)
       return (addr->addr_number <= input->line_number);
 
     case addr_is_last:
-      return test_dollar_EOF(input);
+      return test_eof(input);
 
     default:
       panic(_("INTERNAL ERROR: bad address type"));
@@ -747,9 +769,10 @@ match_address_p(cmd, input)
 }
 
 
-static void do_list P_((void));
+static void do_list P_((int line_len));
 static void
-do_list()
+do_list(line_len)
+     int line_len;
 {
   unsigned char *p = CAST(unsigned char *)line.active;
   countT len = line.length;
@@ -793,7 +816,7 @@ do_list()
 	    }
       }
       olen = o - obuf;
-      if (width+olen >= lcmd_out_line_len && lcmd_out_line_len > 0) {
+      if (width+olen >= line_len && line_len > 0) {
 	  ck_fwrite("\\\n", 1, 2, output_file);
 	  width = 0;
       }
@@ -825,12 +848,12 @@ do_subst(sub)
 		   regs, MAX_BACKREFERENCES))
     return;
   
-  /* We found a match, set the `replaced' flag. */
-  replaced = 1;
-
   if (!sub->replacement && sub->numb <= 1)
     if (regs[0].rm_so == 0 && !sub->global)
       {
+	/* We found a match, set the `replaced' flag. */
+	replaced = 1;
+
 	line.active += regs[0].rm_eo;
 	line.length -= regs[0].rm_eo;
 	line.alloc -= regs[0].rm_eo;
@@ -838,6 +861,9 @@ do_subst(sub)
       }
     else if (regs[0].rm_eo == line.length)
       {
+	/* We found a match, set the `replaced' flag. */
+	replaced = 1;
+
 	line.length = regs[0].rm_so;
 	goto did_subst;
       }
@@ -877,6 +903,9 @@ do_subst(sub)
 	  continue;
 	}
 
+      /* We found a match, set the `replaced' flag. */
+      replaced = 1;
+
       last_end = regs[0].rm_eo;
 
       /* Now expand the replacement string into the output string. */
@@ -979,8 +1008,8 @@ do_subst(sub)
   
   if (sub->print & 2)
     output_line(line.active, line.length, line.chomped, output_file);
-  if (sub->wfile)
-    output_line(line.active, line.length, line.chomped, sub->wfile);
+  if (sub->fp)
+    output_line(line.active, line.length, line.chomped, sub->fp);
 }
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
@@ -1180,20 +1209,30 @@ execute_program(vec, input)
 	      break;
 
 	    case 'l':
-	      do_list();
+	      do_list(cur_cmd->x.int_arg == -1
+		      ? lcmd_out_line_len
+		      : cur_cmd->x.int_arg);
+	      break;
+
+	    case 'L':
+	      fmt(line.active, line.active + line.length,
+		  cur_cmd->x.int_arg == -1
+		  ? lcmd_out_line_len
+		  : cur_cmd->x.int_arg,
+		  output_file);
 	      break;
 
 	    case 'n':
 	      if (!no_default_output)
 		output_line(line.active, line.length, line.chomped, output_file);
-	      if (!read_pattern_space(input, 0))
-		return 0;
+	      if (test_eof(input) || !read_pattern_space(input, vec, 0))
+		return -1;
 	      break;
 
 	    case 'N':
 	      str_append(&line, "\n", 1);
-	      if (!read_pattern_space(input, 1))
-		return 0;
+	      if (test_eof(input) || !read_pattern_space(input, vec, 1))
+		return -1;
 	      break;
 
 	    case 'p':
@@ -1214,13 +1253,33 @@ execute_program(vec, input)
 	      /* Fall through */
 
 	    case 'q':
-	      return cur_cmd->x.exit_status;
+	      return cur_cmd->x.int_arg == -1 ? 0 : cur_cmd->x.int_arg;
 
 	    case 'r':
-	      if (cur_cmd->x.rfile)
+	      if (cur_cmd->x.fname)
 		{
 		  struct append_queue *aq = next_append_slot();
-		  aq->rfile = cur_cmd->x.rfile;
+		  aq->fname = cur_cmd->x.fname;
+		}
+	      break;
+
+	    case 'R':
+	      if (cur_cmd->x.fp && !feof (cur_cmd->x.fp))
+		{
+		  struct append_queue *aq;
+		  size_t buflen;
+		  char *text = NULL;
+		  int result;
+
+		  result = getline (&text, &buflen, cur_cmd->x.fp);
+
+		  if (result != EOF)
+		    {
+		      aq = next_append_slot();
+		      aq->free = 1;
+		      aq->text = text;
+		      aq->textlen = result;
+		    }
 		}
 	      break;
 
@@ -1248,9 +1307,18 @@ execute_program(vec, input)
 	      break;
 
 	    case 'w':
-	      if (cur_cmd->x.wfile)
+	      if (cur_cmd->x.fp)
 		output_line(line.active, line.length,
-			    line.chomped, cur_cmd->x.wfile);
+			    line.chomped, cur_cmd->x.fp);
+	      break;
+
+	    case 'W':
+	      if (cur_cmd->x.fp)
+	        {
+		  char *p = memchr(line.active, '\n', line.length);
+		  output_line(line.active, p ? p - line.active : line.length,
+			      p ? 1 : line.chomped, cur_cmd->x.fp);
+	        }
 	      break;
 
 	    case 'x':
@@ -1259,10 +1327,82 @@ execute_program(vec, input)
 
 	    case 'y':
 	      {
-		unsigned char *p, *e;
-		p = CAST(unsigned char *)line.active;
-		for (e=p+line.length; p<e; ++p)
-		  *p = cur_cmd->x.translate[*p];
+#if defined MBS_SUPPORT && !defined REG_PERL
+               if (MB_CUR_MAX > 1)
+                 {
+                   int idx, prev_idx; /* index in the input line.  */
+                   char **trans;
+                   mbstate_t cur_stat;
+                   memset(&cur_stat, 0, sizeof(mbstate_t));
+                   for (idx = 0; idx < line.length;)
+                     {
+                       int mbclen, i;
+                       mbclen = mbrlen(line.active + idx, line.length - idx,
+                                       &cur_stat);
+                       /* An invalid sequence, or a truncated multibyte
+                          character.  We treat it as a singlebyte character.
+                       */
+                       if (mbclen == (size_t) -1 || mbclen == (size_t) -2
+                           || mbclen == 0)
+                         mbclen = 1;
+
+                       trans = cur_cmd->x.translatemb;
+                       /* `i' indicate i-th translate pair.  */
+                       for (i = 0; trans[2*i] != NULL; i++)
+                         {
+                           if (strncmp(line.active + idx, trans[2*i], mbclen)
+                               == 0)
+                             {
+                               int move_remain_buffer = 0;
+                               int trans_len = strlen(trans[2*i+1]);
+
+                               if (mbclen < trans_len)
+                                 {
+                                   int new_len;
+                                   new_len = line.length + 1 + trans_len - mbclen;
+                                   /* We must extend the line buffer.  */
+                                   if (line.alloc < new_len)
+                                     {
+                                       /* And we must resize the buffer.  */
+                                       resize_line(&line, new_len);
+                                     }
+                                   move_remain_buffer = 1;
+                                 }
+                               else if (mbclen > trans_len)
+                                 {
+                                   /* We must truncate the line buffer.  */
+                                   move_remain_buffer = 1;
+                                 }
+                               prev_idx = idx;
+                               if (move_remain_buffer)
+                                 {
+                                   int move_len, move_offset;
+                                   char *move_from, *move_to;
+                                   /* Move the remaining with \0.  */
+                                   move_from = line.active + idx + mbclen;
+                                   move_to = line.active + idx + trans_len;
+                                   move_len = line.length + 1 - idx - mbclen;
+                                   move_offset = trans_len - mbclen;
+                                   memmove(move_to, move_from, move_len);
+                                   line.length += move_offset;
+                                   idx += move_offset;
+                                 }
+                               strncpy(line.active + prev_idx, trans[2*i+1],
+                                       trans_len);
+                               break;
+                             }
+                         }
+                       idx += mbclen;
+                     }
+                 }
+               else
+#endif /* MBS_SUPPORT */
+                 {
+                   unsigned char *p, *e;
+                   p = CAST(unsigned char *)line.active;
+                   for (e=p+line.length; p<e; ++p)
+                     *p = cur_cmd->x.translate[*p];
+                 }
 	      }
 	      break;
 
@@ -1277,17 +1417,20 @@ execute_program(vec, input)
 	}
 
 #ifdef EXPERIMENTAL_DASH_N_OPTIMIZATION
-/* If our top-level program consists solely of commands with addr_is_num
- * addresses then once we past the last mentioned line we should be able
- * to quit if no_default_output is true, or otherwise quickly copy input
- * to output.  Now whether this optimization is a win or not depends
- * on how cheaply we can implement this for the cases where it doesn't
- * help, as compared against how much time is saved.
- * One semantic difference (which I think is an improvement) is
- * that *this* version will terminate after printing line two
- * in the script "yes | sed -n 2p".
- */
-      else
+      /* If our top-level program consists solely of commands with
+       * addr_is_num addresses then once we past the last mentioned
+       * line we should be able to quit if no_default_output is true,
+       * or otherwise quickly copy input to output.  Now whether this
+       * optimization is a win or not depends on how cheaply we can
+       * implement this for the cases where it doesn't help, as
+       * compared against how much time is saved.  One semantic
+       * difference (which I think is an improvement) is that *this*
+       * version will terminate after printing line two in the script
+       * "yes | sed -n 2p". 
+       *
+       * Don't use this when in-place editing is active, because line
+       * numbers restart each time then. */
+      else if (output_file == stdout)
 	{
 	  /* can we ever match again? */
 	  if (cur_cmd->a1->addr_type == addr_is_num &&
@@ -1353,7 +1496,7 @@ process_files(the_program, argv)
   input.fp = NULL;
 
   status = EXIT_SUCCESS;
-  while (read_pattern_space(&input, 0))
+  while (read_pattern_space(&input, the_program, 0))
     {
       status = execute_program(the_program, &input);
       if (!no_default_output)
